package net.pingex.dcf.core;

import net.pingex.dcf.util.DiscordInteractionsUtil;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import sx.blah.discord.api.ClientBuilder;
import sx.blah.discord.api.IDiscordClient;
import sx.blah.discord.util.DiscordException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Storage of all active connection to Discord gateway
 */
public class GatewayConnectionsManager
{
	/**
	 * Main datastore
	 */
	private Set<IDiscordClient> connectionsDatastore;

	/**
	 * List of registered listeners for each connection, as we can't access registered listeners in IDC dispatcher.
	 */
	private Map<IDiscordClient, Set<Object>> registeredListeners;

	/**
	 * Singleton unique instance
	 */
	private static final GatewayConnectionsManager INSTANCE = new GatewayConnectionsManager();

	/**
	 * Logger
	 */
	private static final Logger LOGGER = LogManager.getLogger(GatewayConnectionsManager.class);

	/**
	 * Get singleton instance
	 */
	public static GatewayConnectionsManager getInstance()
	{
		return INSTANCE;
	}

	/**
	 * Private constructor, to enforce singleton
	 */
	private GatewayConnectionsManager()
	{
		connectionsDatastore = new HashSet<>();
		registeredListeners = new HashMap<>();
	}

	/**
	 * Registers a connection and login automatically
	 * @param builder Filled and builder ready for log in
	 */
	public void registerConnection(ClientBuilder builder)
	{
		LOGGER.info("Registering new connection");
		builder.withReconnects(); // Makes the connections auto-reconnects on dropped connection.
		try
		{
			IDiscordClient builtConnection = builder.login();
			connectionsDatastore.add(builtConnection);
			builtConnection.getDispatcher().registerListener(CoreEventsHandler.class); // Register the core event handler independently from the events package
			//updateListeners(builtConnection, null); // TODO: EventRegistry
		}
		catch (DiscordException e)
		{
			LOGGER.warn("Failed to login to Discord.", e);
		}
	}

	/**
	 * Logout and unregister specified connection
	 * @param client Target connection
	 */
	public void unregisterConnection(IDiscordClient client)
	{
		LOGGER.info("Unregistering connection with user #" + client.getOurUser().getID());

		DiscordInteractionsUtil.disconnect(client);
		connectionsDatastore.remove(client);
		registeredListeners.remove(client);
	}

	/**
	 * Update listeners for target connection
	 * @param target Target conection, must be registered
	 * @param refListeners Reference listeners list
	 */
	public void updateListeners(IDiscordClient target, Set<Object> refListeners)
	{
		LOGGER.debug("Updating listeners for target " + target.getOurUser().getID());

		if(!connectionsDatastore.contains(target)) return;
		if(!registeredListeners.containsKey(target)) registeredListeners.put(target, new HashSet<>());

		Set<Object> currentListeners = registeredListeners.get(target);

		// Case 1: item is not registered in IDC
		long toRegister = refListeners.stream().filter(item -> !currentListeners.contains(item)).count();
		refListeners.stream().filter(item -> !currentListeners.contains(item)).forEach(item ->
		{
			target.getDispatcher().registerListener(item);
			currentListeners.add(item);
		});

		// Case 2: item is registered, but shouldn't be
		long toUnregister = currentListeners.stream().filter(item -> !refListeners.contains(item)).count();
		currentListeners.stream().filter(item -> !refListeners.contains(item)).forEach(item ->
		{
			target.getDispatcher().unregisterListener(item);
			currentListeners.remove(item);
		});

		LOGGER.debug("Registered " + toRegister + " listeners and unregistered " + toUnregister + " listeners.");
	}

	/**
	 * Update all connections listeners
	 * @param refListeners Reference listeners list
	 */
	public void updateAllListeners(Set<Object> refListeners)
	{
		for(IDiscordClient i : connectionsDatastore) updateListeners(i, refListeners);
	}

	/**
	 * Reconnect a disconnected gateway connection
	 * @param target Disconnected connection
	 */
	public void reconnect(IDiscordClient target)
	{
		if(!connectionsDatastore.contains(target)) return;

		try
		{
			target.login();
		}
		catch (DiscordException e)
		{
			LOGGER.warn("User #" + target.getOurUser().getID() + " failed to reconnect to the gateway.", e);
		}
	}
}
